/*  Lexical Analyzer of MIKC language
 *  Copyright (C) 2003 Michal Conos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <lex.h>
#include <compile.h>

#ifndef NOGFX
#	include "gtk/gtk-draw.h"
#endif

#ifndef PI
#define PI 3.141592653589
#endif

#define MAX_MEM           1024
#define MAX_STACK         1024;

struct MemoryElement {
	char   *strval;
	double  numval;
};

struct StackElement {
	char   *strval;
	double  numval;
	struct MemoryElement * memptr;  /* ASSIGN */
};


static struct MemoryElement varmem [MAX_MEM];
static struct StackElement  stack  [MAX_MEM];
static struct Instruction   progmem[MAX_PROG_LEN];

static int SP = 0;
static int PC = 0;

static int len;

LexAnalyzer *lex;


static double x=0;
static double y=0;
static double a=0;

static double ox = 0;
static double oy = 0;

#define WIN_WIDTH   640
#define WIN_HEIGHT  480

#ifndef NOGFX
static int sx = WIN_WIDTH  >> 1;
static int sy = WIN_HEIGHT >> 1;
#endif

void
forward( double num )
{
	x = ox + num * cos(a*PI/180);
	y = oy + num * sin(a*PI/180);
	
	printf("DRAW: [%.2f, %.2f] -> [%.2f, %.2f] (%.2f)\n", ox, oy, x, y, num );
#ifndef NOGFX
	draw_line( sx + (int)floor(ox), sy - (int)floor(oy), sx + (int)floor(x), sy - (int)floor(y) );
#endif

	ox = x;
	oy = y;
}

void 
back( double num )
{
	x = ox - num * cos(a*PI/180);
	y = oy - num * sin(a*PI/180);

	printf("DRAW: [%.2f, %.2f] -> [%.2f, %.2f] (%.2f)\n", ox, oy, x, y, num );
#ifndef NOGFX
	draw_line( sx + (int)floor(ox), sy - (int)floor(oy), sx + (int)floor(x), sy - (int)floor(y) );
#endif
	
	ox = x;
	oy = y;
}

void
left( double num )
{
	a+=num;
#ifdef NOGFX
	printf("DRAW: alpha=[%.2f]\n", a );
#else
#endif
}

void 
right( double num )
{
	a-=num;
#ifdef NOGFX
	printf("DRAW: alpha=[%.2f]\n", a );
#else
#endif
}

void 
penup()
{
	printf("penup not  implemented yet\n");
}

void 
pendown()
{
	printf("pendown not  implemented yet\n");
}


static void
read_binary( const char *filename )
{
	FILE *fp;
	u_int32_t buf;

	fp = fopen( filename, "rb" );
	if ( !fp ) {
		perror( __FILE__ ":fopen()" );
		exit( EXIT_FAILURE );
	}

	while ( !feof(fp ) ) {
		fread( &buf, sizeof(u_int32_t), 1, fp );
		progmem[PC] . opcode  = (opcode_t) buf;
		fread( &buf, sizeof(u_int32_t), 1, fp );
		progmem[PC] . operand = (int)      buf;
		PC++;
	}

	len = --PC;

	fprintf(stderr,__FILE__ ": PC=0x%x\n", PC);
	PC = 0;
	
	fclose( fp );
}

/* GO ... */
void
GO(void)
{
	struct Instruction instr;
	int oper;
	double l, r;
	struct StackElement stl, str;
	struct MemoryElement *mel;
	int li, ri;
	
	for ( ;; ) {

		if ( PC >= len ) break;
		instr = progmem[PC++];
		oper  = instr.operand;
	//	printf("oper=%d\n", oper);

		switch ( instr.opcode ) {
			case TA:     /* TA addr */
				++SP;
				stack[SP].numval = varmem[oper].numval;
				stack[SP].memptr = &varmem[oper];
				printf("TA %d, SP=%d, stack[SP].numval=%.2f adr=%p\n", oper, SP, stack[SP].numval, stack[SP].memptr );
				break;
			case TC:     /* TC const */
				stack[++SP].numval = oper;	
				stack[SP].memptr = NULL;
				printf("TC %d, SP=%d, stack[SP].numval=%.2f\n", oper, SP, stack[SP].numval );
				break;
			case BOP:
				str = stack[SP--];
				stl = stack[SP--];
				r = str.numval;
				l = stl.numval;
				ri = (int) floor(r);
				li = (int) floor(l);
				
				switch ( oper ) {
					case PLUS:
						stack[++SP].numval = l + r;
						printf("BOP + , SP=%d, stack[SP].numval=%.2f\n", SP, stack[SP].numval );
						break;
					case MINUS:
						stack[++SP].numval = l - r;
						break;
					case MULTIPLY:
						stack[++SP].numval = l * r;
						break;
					case DIVIDE:
						stack[++SP].numval = l / r;
						break;
					case MODULO:
						stack[++SP].numval = (double)(li % ri);
						break;
					case BIT_AND:
						stack[++SP].numval = (double)(li & ri);
						break;
					case BIT_OR:
						stack[++SP].numval = (double)(li | ri);
						break;
					case BIT_XOR:
						stack[++SP].numval = (double)(li ^ ri);
						break;
					case LOGIC_AND:
						stack[++SP].numval = (double)(l && r);
						break;
					case LOGIC_OR:
						stack[++SP].numval = (double)(l || r);
						break;
					case LOGIC_XOR:
						stack[++SP].numval = (double)((!l && r) || (l && !r));
						break;
					case LEFT_SHIFT:
						stack[++SP].numval = (double)(li << ri);
						break;
					case RIGHT_SHIFT:
						stack[++SP].numval = (double)(li >> ri);
						break;
					case EQ:
						stack[++SP].numval = (double)(l == r);
						break;
					case GE:
						stack[++SP].numval = (double)(l >= r);
						break;
					case LE:
						stack[++SP].numval = (double)(l <= r);
						break;
					case LT:
						stack[++SP].numval = (double)(l < r);
						break;
					case GT:
						stack[++SP].numval = (double)(l > r);
						break;
					case ASSIGN:
						if ( stl . memptr == NULL ) {
							fprintf( stderr, __FILE__ ": You're attempting to assign to non-variable expression!\n" );
							exit (EXIT_FAILURE );
						}
						printf("stl.memptr=%p &(varmem[0].numval)=%p &(varmem[0])=%p\n", stl.memptr, &(varmem[0].numval), &(varmem[0]));
						mel = stl.memptr;
						mel->numval = r;
						printf(" mel.numval=%p r=%.2f\n", &(mel->numval), r);
						stack[++SP].numval = r;
						printf("varmem[0].numval=%.2f\n", varmem[0].numval);
						printf("BOP = , SP=%d, stack[SP].numval=%.2f adr=%p\n", SP, stack[SP].numval, stack[SP].memptr );
						break;
					default:
						fprintf( stderr, __FILE__ ": Unknown binary operator (%d) !\n", oper );
						exit( EXIT_FAILURE );
				}
				break;
			case UNOP:
				switch ( oper ) {
					case MINUS:
						stack[SP].numval = - stack[SP].numval;
						break;
					case BIT_NEG:
						stack[SP].numval = (double)( ~ (int) floor ( stack[SP].numval ) );
						break;
					case LOGIC_NOT:
						stack[SP].numval = !stack[SP].numval;
						break;
					default:
						fprintf( stderr, __FILE__ ": Unknown unary operator (%d) !\n", oper );
						exit( EXIT_FAILURE );

				}
				break;
			case JZ:
				if ( ! stack[SP--].numval ) PC = oper;
				break;
			case JMP:
				PC = oper;
				break;
			case WRT:
				printf("PRINT: %.2f\n", stack[SP--].numval );
				break;
			case FD:
				printf("FD , SP=%d, stack[SP].numval=%.2f\n", SP, stack[SP].numval );
				forward( stack[SP--].numval );
				break;
			case BK:
				back( stack[SP--].numval );
				break;
			case LFT:
				left( stack[SP--].numval );
				break;
			case RGT:
				right( stack[SP--].numval );
				break;
			case PU:
				penup();
				break;
			case PD:
				pendown();
				break;
			default:
				fprintf( stderr, __FILE__ ": Unknown instruction (%d)!\n", instr.opcode );
				exit( EXIT_FAILURE );
		}
	}
}


int
main( int argc, char *argv[] )
{

	lex = new LexAnalyzer( NULL );
	
	read_binary( "bin.out" );
	dissassemble( progmem, len, stdout );
	
#ifndef NOGFX
	open_window( argc, argv, WIN_WIDTH, WIN_HEIGHT );
#endif
	
	GO();

#ifndef NOGFX
    call_gtk_main();
#endif

	return EXIT_SUCCESS;
}
