/* Syntakticky analyzer LL(1) gramatik. 
 * Prijima pravidla ve tvaru:
 * 	A > BC | e    ( A se prepisuje na BC nebo e )
 * 	B > aB|bB|cB|e
 * 	C > 0C|1C|2C|e
 *
 * 	,kde [A-Z] jsou neterminaly, zbytek jsou terminaly, krome e, ktere reprezentuje prazdne slovo
 *
 * 	Hleda FIRST a FOLLOW fce pro neterminaly 
 * 	Vytvori SELECT tabulku a vygeneruje funkce rekurzivniho sestupu pro danou gramatiku
 *
 * 	Cte ze stdin.
 *
 * 	Copyright (C) 2002, 2003 Michal Conos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 */


#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_STRING_H
#  include <string.h>
#endif

#include <utility.h>
#include <ll.h>

#define MAX_RULES  1024
#define MAX_SYMBOLS 256

/* Nactena pravidla */
static char *rules[MAX_SYMBOLS][MAX_RULES];

/* Tady jsou ulozeny symboly */
static int left_side_symbols[MAX_SYMBOLS];

/* Tady je reference count, kvuli zacykleni u _follow fce */
static int ref_count[MAX_SYMBOLS];

static int lsyms = 0;

#define INIT_SIZE 64

static char *FIRST[MAX_SYMBOLS];
static char *FOLLOW[MAX_SYMBOLS];
int act_first_size[MAX_SYMBOLS];
int act_follow_size[MAX_SYMBOLS];

/* Prvni symbol, ktery mame */
int *Start_Symbol = &left_side_symbols[0];

static int very_verbose = 0;

#define NON_TERM(x) ( (x)>='A' && (x)<='Z' ) 
#define TERM(x)    ( ! NON_TERM(x) )
	
/* Souhrn fci, ktere zaviseji na datove reprezentaci pravidel, jsou oznaceny jako static  */
static int
search_lsym( int sym )
{
	int i;
	for ( i = 0; i < lsyms; i++ ) 
		if ( left_side_symbols[i] == sym )
			return 1;
		
	return 0;
}

static int
search_rule( int sym, char *rule )
{
	int cnt=0;
	char *s;

	while ( ( s = rules[ sym ][ cnt++ ] ) != NULL ) {
		if ( !strcmp( s, rule ) )  {
			return 1;
		}
	}
	return 0;
}

static int
check_empty_word( int x )
{
	return ( (char)x == 'e' ) ? 1 : 0 ;
}

static int
expands_empty_word( int x ) 
{
	return  ( !(strchr(FIRST[x], 'e')) ) ? 0 :  1 ;
}

static int
contains_empty_word( char *str ) 
{
	return  ( !(strchr(str, 'e')) ) ? 0 :  1 ;
}

/* hleda 'symbol' v 'rule' a najde-li jej, vrati nasledujici v 'next', 
 * navrat 1 pri uspechu jinak 0*/
static int
get_next_symbol( char *rule, int symbol, int *next )
{
	char x = (char)symbol, *s;

	if (!rule) return 0;
	
	if ( ( s = strchr( rule, x ) ) != NULL ) {
		s++;
		*next = *s;
		return *s; 
	} else {
		return 0;
	}
}

static int
add_symbol_if_neccessary( char *f, char s, int len )
{
	if ( !strchr( f, s ) ) {
		f[len] = s;
		return 1;
	}
	return 0;
}

static void
add_string_if_neccessary( char *dest, char *src ) 
{
	int i,j;

	if (!dest || !src) return;

	j = strlen(dest);
	for (i=0; i<strlen(src); i++){
		if ( add_symbol_if_neccessary( dest, src[i], j )) j++;
	}
	
}

/* --------------------------------------------------------------------------*/


/* Cte pravidla ze stream-u 
 * Odstrani duplicitni pravidla( kazde pravidlo se vyskytuje jen jednou ) */
void
read_rules( FILE *fp )
{
	char *line,*running,*left,*right, *token;
	char *delim = ">", *p;
	int N,line_num=0, cnt,i;

	/* init */
	for ( i=0; i<MAX_SYMBOLS; i++ ) {
		rules[i][0]=NULL;
		FIRST[i] = xcalloc( INIT_SIZE, 1 );
		FOLLOW[i] = xcalloc( INIT_SIZE, 1 );
		act_follow_size[i] = INIT_SIZE;
		act_first_size[i]  = INIT_SIZE;
	}

	memset( left_side_symbols, 0, MAX_SYMBOLS );
	
	
	while ( line=NULL, N=0, getline( &line, &N, fp )!=-1 ) {
		line_num++;
		/* odstran '\n' */
		p = strchr(line, '\0');
		p--;
		*p='\0';
		running = xstrdup(line);

		if ( line[0]=='#' ) continue; /* Comment */
		if ( line[0]=='\n' ) continue; /* New line */

		
		/* rozdeli pravidlo na levou a pravou cast */
		left = strsep( &running, delim ); 
		right = strsep( &running, delim );
		
		if ( !left || !right ) {
			fprintf(stderr," Error on line #%d!\n", line_num);
			continue;
		}
		
		/* smazeme mezery */
		left=delete_spaces(left);
		right=delete_spaces(right);
		
		free( running );
		
		
		if ( strlen(left) > 1 ) {
			fprintf(stderr, "Error on line #%d - too many characters on the left side of rule!\n", line_num);
		}

		/* check, ze na leve strane je neterminal */
		if ( ! NON_TERM( left[0] ) ) {
			fprintf(stderr, "Error on line #%d - On the left there must be only non-terminal symbols(A-Z)\n", line_num);
			continue;
		}
		
		cnt = -1;
		while ( rules[ (int)left[0] ][++cnt] ) ;

		if ( cnt >= MAX_RULES ) {
			fprintf(stderr,"Sorry, space for storing rules for this symbol exhausted (MAX_RULES=%d)\n", MAX_RULES);
			continue;
		}
		
	
		if ( lsyms > MAX_SYMBOLS ) {
			fprintf(stderr,"Sorry, space for storing symbols is exhausted (MAX_SYMBOLS=%d)\n", MAX_SYMBOLS);
			return;
		}
		
		if ( !search_lsym( (int) left[0] ) ) {
			left_side_symbols[lsyms++] = (int) left[0];
		}
		
		/* stage 2 : koukeneme se po na prave strane '|' jakozto oznacovaci variant */
		running = xstrdup(right);

		while ( ( token = strsep( &running, "|" ) ) != NULL ) {
			token = delete_spaces( token );
			if ( !search_rule( (int)left[0], token ) ) {
				ASSIGN( rules[ (int)left[0] ][cnt], token );
				cnt++;
				rules[ (int)left[0] ][cnt]=NULL;
			}
		}
		
		free( running );
	}
}

/* --------------------------------------------------------------------------*/
/* Vytiskne vsechna nactena pravidla */
void
print_rules( void )
{
	int i,j, c=0;
	int lsym;

	printf("\nRead rules:\n\n");
	for ( i = 0; i<lsyms; i++ ) {
		lsym = left_side_symbols[i];
		j = 0;
		while ( rules[ (int)lsym ][ j++ ] ) {
			printf("(%d) %c -> %s\n", ++c, lsym, rules[ (int)lsym][j-1]);
		}
	}
}

/* ---- TADY ZACINAJI OPRAVDOVE ALGORITMY ---- */


/* ---- FIRST FUNKCE ------------------------- */

static void 
_first( int lsym, int x, int follow_mode )
{
	int cnt=0;
	char *rule;	

	if ( NON_TERM( x ) ) {
		if ( !rules[x][0] ) {
			fprintf(stderr, "No rule specified for this symbol (%c)!\n", x);
			exit( EXIT_FAILURE );
		}
		while ( ( rule = rules[x][cnt++] ) != NULL ) {
			_first( lsym, (int) rule[0], follow_mode );
		}
	} else {
		/* nasli jsme terminal */
		char *dummy;
		int len;  
		printf("lsym=%c\n", lsym);
		if ( follow_mode ) dummy = FOLLOW[lsym]; else dummy = FIRST[lsym];
		len = strlen(dummy);

		if (very_verbose)
		printf("\t+ first found: %c (%c)\n", x, (char)lsym);
		if ( len  > act_first_size[lsym] ) {
			act_first_size[lsym] <<= 1;
			dummy = xrealloc( dummy, act_first_size[lsym] );
			dummy[act_first_size[lsym]-1]='\0';
		}

		add_symbol_if_neccessary( dummy, x, len );
	}
}

/* Vytvori tabulku FIRST pro vsechny symboly */
void
first(void)
{
	int i,x, cnt, next_symbol,lsym,j,q;
	char *rule;
	
	/* stage 1: najdi first termy pro kazdy NON_TERM symbol */
	for ( i = 0; i < lsyms; i++ ) {
		x = left_side_symbols[i];
		if(very_verbose)
		printf("\n\nSearching for %c:\n", (char)x);
		_first( x, x, 0 );
	}

	/* stage 2: A -> BC, e \in FIRST(B), pak jeste FIRST(C) */
	for ( i = 0; i < lsyms; i++ ) {
		if ( expands_empty_word( x = left_side_symbols[i] ) ) {
			if(very_verbose)
			printf("*Symbol %c could expand to an empty word\n", (char)x);

			/* Hledej v pravidlech symbol, ktery expanduje na 'e', zjisti pro nasledujici term first fci */

			for ( j = 0; j<lsyms; j++ ) {
				lsym = left_side_symbols[j];
				cnt = 0;
				while  ( ( rule=rules[lsym][cnt++] ) != NULL ) {
					for ( q=0; q<strlen(rule); q++ ) {
						 if ( ! expands_empty_word( rule[q] ) ) 
							 /* Jestlize nektere z predchazejicich neexpanduje na 'e', konec */
							 break;
						 if ( rule[q] == (char)x ) {
							  next_symbol = rule[q+1];
							  if ( (char)next_symbol != '\0' ) {
									_first( lsym, next_symbol, 0 ); 	
							  }
						 }
					}
				}
			}

		}
	}

}

/* Tiskne FIRST_TABLE pro vsechny symboly */
void
print_first(void)
{
	int i;
	printf("\n\nFIRST TABLE:\n");
	for ( i = 0; i < lsyms; i++ ) {
		printf("(%c) : %s\n", (char)left_side_symbols[i], FIRST[ left_side_symbols[i] ]);
	}
}

/* ---- FOLLOW FUNKCE ------------------------- */

static char *
find_first( int sym )
{
	char *f;
	if ( NON_TERM( sym ) ) {
		return FIRST[sym];
	} else {
		f = xcalloc( 2, 1 );
		f[0] = sym;
		return f;
	}
}

/* Tohle nam vynuluje reference, ktere tam mohla zanest _follow() */
static void
zero_all_references( void )
{
	int x, i;
	for ( i=0; i<lsyms; i++ ) {
		x = left_side_symbols[i];
		ref_count[x] = 0;
	}
}
static void
__follow( int lsym, int Nterm )
{
	int i, x, cnt, next,j;
	char *rule;
	char *first;
	

	if(very_verbose)
	printf("\n\nSearching %c (examinig: %c[%d]):\n", (char)lsym, (char)Nterm, ref_count[Nterm]);
	if (ref_count[Nterm] ++) 
		/* Zacarovany kruh, tady uz se nic noveho nedozvime */
		return;
	
	
	/* projdi pravidla od vsech symbolu */
	for ( i=0; i < lsyms; i++ ) {
		x = left_side_symbols[i];
		cnt = 0;

		while ( (rule = rules[x][cnt++]) != NULL ) {
			for ( j=0; j<strlen(rule); j++)  {
				if ( rule[j] == (char)Nterm ) {
				/* nasli jsme neterminal v pravidle */
				if(very_verbose)
				printf("%c -> %s\n", (char)x, rule);
				next = (int)rule[j+1];	
					if ( (char)next == '\0' && Nterm != x ) {
						/* byl na poslednim miste, potrebujeme znat FOLLOW(x) */
						__follow(lsym, x);
					} else {
						/* neni na poslednim miste, chceme FIRST(next) */
						first =  find_first( next );
						if ( strlen(FOLLOW[lsym]) + strlen(first) + 1 >= act_follow_size[lsym] ) {
								act_follow_size[lsym] <<= 1;
								FOLLOW[lsym] = xrealloc( FOLLOW[lsym], act_follow_size[lsym] );
						}

						add_string_if_neccessary( FOLLOW[lsym], first );
						if(very_verbose)
						printf("\tfirst=%s\n", first);
						if ( contains_empty_word( first ) && Nterm != x) {
								/* Hmmm, first obsahoval 'e', potrebujeme znat FIRST(x) */
								__follow(lsym, x);
						}
					}
				}	
			}
		}
	}
}

void
follow( void )
{
	int i, x;
	
	if(very_verbose)
	printf("*** FOLLOW (%d):\n", lsyms);
	
	/* vsechny neterminaly na 'e' (krome start_symbolu) */
	for ( i=1; i<lsyms; i++) {
		x = left_side_symbols[i];
		add_symbol_if_neccessary( FOLLOW[x], 'e', 0 );
	}

	/* vsechny non-terminaly na prave strane pravidel */
	for ( i=0; i<lsyms; i++ ) {
		x = left_side_symbols[i];
		/* Musime vynulovat reference count, co tam zbyly z minuleho volani _follow() */
		zero_all_references();
		__follow(x, x);
	}
}
/* Tiskne FOLLOW_TABLE pro vsechny symboly */
void
print_follow(void)
{
	int i;

	printf("\n\nFOLLOW TABLE:\n");
	for ( i = 0; i < lsyms; i++ ) {
		printf("(%c) : %s\n", (char)left_side_symbols[i], FOLLOW[ left_side_symbols[i] ]);
	}
}

/* ------------ Funkce rekurzivniho sestupu -------------- */
void
generate_output( void )
{
	int i,j, c=0,k;
	int lsym;
	char *first_str;
	char *rule;
	char *select_str;
	char *follow_str;
	char *res;

	for ( i = 0; i<lsyms; i++ ) {
		lsym = left_side_symbols[i];
		printf("void function_%c(void);\n", (char)lsym);
	}

	printf("\n/* Select rules: */\n\n");

	/* projdeme vsechna pravidla a najdeme pro ne SELECT(A,rule) */
	for ( i = 0; i<lsyms; i++ ) {
		lsym = left_side_symbols[i];
		j = 0;
		
		printf("\nvoid function_%c( void )\n", (char)lsym);
		printf("{\n");
		printf("	switch (input) {\n");
		
		while ( (rule = rules[ (int)lsym ][ j++ ]) != NULL ) {
			first_str = find_first( (int)rule[0] );
			if ( contains_empty_word( first_str ) ) {
				/* obsahuje-li FIRST(rule) 'e', musime znat jeste FOLLOW(A) */
				follow_str = FOLLOW[lsym];
				select_str = xcalloc( strlen(first_str) + strlen(follow_str) + 1, 1 );
				add_string_if_neccessary(select_str, first_str);
				add_string_if_neccessary(select_str, follow_str);
				res = xstrdup(select_str);
				free(select_str);
			} else {
				res = xstrdup(first_str);
			}
			/* projed SELECT symboly*/
			for ( k=0; k < strlen(res); k++ ) {
				if ( check_empty_word(res[k]) ) {
					printf("\t\tcase EOF:\n");
					printf("\t\tcase '\\n':\n");
//					printf("\t\t\tprintf(\"EOF\\n\");\n");
				} else {
					printf("\t\tcase '%c':\n", res[k]);
				}
				if (k==strlen(res)-1){
					printf("\t\t/* (%d) %c -> %s\t [%s] */\n", ++c, lsym, rule, res);
					printf("\t\t\tprintf(\" (%d) %c -> %s\t [%s] \\n\");\n", c, lsym, rule, res);
				}
			}

			/* projed pravidlo a sestav poradi funkci */
			for ( k=0; k < strlen(rule); k++) {
				if ( NON_TERM( rule[k] ) ) {
					printf("\t\t\tfunction_%c();\n", rule[k]);
				} else {
					if ( !check_empty_word( (int) rule[k] ) ) {
						printf("\t\t\tcompare( (int)'%c' );\n", rule[k]);
					}
				}
			}
			printf("\t\t\tbreak;\n");
		}
		printf("		default:\n");
		printf("			fprintf(stderr, \"Input parse error!\\n\");\n");
		printf("			exit( EXIT_FAILURE );\n");
		printf("	}\n");
	
		printf("}\n");
	}

}


int 
main( int argc, char *argv[] ) 
{
	read_rules( stdin );
	puts(prolog);
	printf("/*\n");
	print_rules();
	first();
	follow();
	print_first();
	print_follow();
	printf("*/\n");
	generate_output();
	puts(epilog1);
	printf("\tfunction_%c();\n", (char)left_side_symbols[0]);
	puts(epilog2);
	return EXIT_SUCCESS;
}


