/*  Expression operators of MIKC language
 *  Copyright (C) 2003 Michal Conos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <math.h>
#include <lex.h>
#include <node.h>
#include <mystring.h>

#define EPS 0.001

VarValue::VarValue( double num )
{
	numval = num;
}

VarValue::VarValue( char *s )
{
	if ( !s  ){
		fprintf( stderr,__FILE__ ": Attemt to construct NULL VarValue()!!!\n" );
		exit(EXIT_FAILURE);
	}
	strval = s;
}


BinOp::BinOp( lexsym_t oper, Node *left, Node *right )
{
	op = oper;
	switch ( op ) {
		case PLUS:
		case MINUS:
		case MULTIPLY:
		case DIVIDE:
		case MODULO:
		case ASSIGN:
		case LEFT_SHIFT:
		case RIGHT_SHIFT:
		case BIT_AND:
		case BIT_OR:
		case BIT_XOR:
		case EQ:
		case GT:
		case LT:
		case GE:
		case LE:
		case LOGIC_AND:
		case LOGIC_OR:
		case LOGIC_XOR:
			l  = left; r = right;
			break;
		default:
			fprintf(stderr, __FILE__ ": Unknown Operator (%d) !\n", oper );
			exit( EXIT_FAILURE );
	}
}

BinOp::~BinOp() 
{
}

/* gets value of the node */
VarValue * 
BinOp::getValue()
{
	VarValue *val;
	lexsym_t node_type;
	
	node_type  = l -> getType();
	
	if ( l -> getType() != r -> getType() ) {
		fprintf( stderr, __FILE__ ": You cannot operand with different types! E.g.: \"string\"+3 is bad!\n");
		exit( EXIT_FAILURE );
	}
	
	switch ( node_type ) {
		case NUMB:
			val = make_on_numbers( op, l, r );
			break;
		case DOUBLE_QUOTE:
			val = make_on_strings( op, l, r );
			break;
		default:
			fprintf( stderr, __FILE__ ": Ooops, Unknown operand type (%d)!\n", node_type );
			exit( EXIT_FAILURE );
	}

	return val;
}

/* Binary Opearations on numbers */
VarValue *
BinOp::make_on_numbers( lexsym_t op, Node *left, Node *right )
{
	double lv = left -> getValue() -> numval;
	double rv = right -> getValue() -> numval;
	int lvi = (int)floor(lv);
	int rvi = (int)floor(lv);

	switch ( op ) {
		case PLUS:
			return new VarValue( lv + rv );
		case MINUS:
			return new VarValue( lv - rv );
		case MULTIPLY:
			return new VarValue( lv * rv );
		case DIVIDE:
			return new VarValue( lv / rv );
		case ASSIGN:
			return new VarValue( rv );
		case LEFT_SHIFT:
			return new VarValue( (double)( lvi << rvi ) );
		case RIGHT_SHIFT:
			return new VarValue( (double)( lvi >> rvi ) );
		case BIT_AND:
			return new VarValue( (double)( lvi & rvi ) );
		case BIT_OR:
			return new VarValue( (double)( lvi | rvi ) );
		case BIT_XOR:
			return new VarValue( (double)( lvi ^ rvi ) );
		case EQ:
			return new VarValue( (double) ( (fabs(lv-rv)<EPS) ? 1.0 : 0.0 ) );
		case GT:
			return new VarValue( (double) (lv>rv) );
		case LT:
			return new VarValue( (double) (lv<rv) );
		case GE:
			return new VarValue( (double) (lv>=rv) );
		case LE:
			return new VarValue( (double) (lv<=rv) );
		case LOGIC_AND:
			return new VarValue( (double) (lv && rv ) );
		case LOGIC_OR:
			return new VarValue( (double) (lv || rv ) );
		case LOGIC_XOR:
			return new VarValue( (double) ((lv && !rv)||(!lv && rv)) );
		default:
			fprintf(stderr,__FILE__ ":Bad binary operator: %d!\n", op);
			exit( EXIT_FAILURE );
	}

}

/* Binary Opeartions on strings */
VarValue *
BinOp::make_on_strings( lexsym_t op, Node *left, Node *right )
{
	char * lv = left -> getValue() -> strval;
	char * rv = right -> getValue() -> strval;

	switch (op) {
		case PLUS:
			return new VarValue( add_string( lv, rv ) );
		case ASSIGN:
			return new VarValue( assign_string( rv ) );
		default:
			fprintf(stderr,__FILE__ ":Bad binary operator on strings: %d!\n", op);
			exit( EXIT_FAILURE );
			
	}
	
}

/* Unary Operator */
UnOp::UnOp( lexsym_t oper, Node *node )
{
	op = oper;
	switch (op) {
		case MINUS:
		case BIT_NEG:
		case LOGIC_NOT:
			n = node;
			break;
		default:
			fprintf(stderr,__FILE__ ": Ivalid Unary Operator (%d)!\n", op);
			exit(EXIT_FAILURE);
	}
}

UnOp::~UnOp()
{
}

/* Get Unary Value */
VarValue *
UnOp::getValue()
{
	VarValue *val;
	
	switch ( n -> getType() ) {
		case NUMB:
			val = make_on_numbers( op, n );
			break;
		default:
			fprintf( stderr, __FILE__ ": Ooops, Unknown operand type (%d)!\n", n -> getType() );
			exit( EXIT_FAILURE );
	}

	return val;

}

/* unary operator can deal only with numbers! */
VarValue *
UnOp::make_on_numbers( lexsym_t op, Node *node )
{
	double v = node -> getValue() -> numval;
	int vi = (int)floor(v);

	switch ( op ) {
		case BIT_NEG:
			return new VarValue( (double) ( ~vi ) );
		case LOGIC_NOT:
			return new VarValue( (double) ( !v ) );
		case MINUS:
			return new VarValue( -v );
		default:
			fprintf( stderr, __FILE__ ":Unknown Unary operator : (%d) !!!\n", op);
			exit( EXIT_FAILURE );
	}
}


/* Leaf structure of the expression tree */
DoubleVarOrConst::DoubleVarOrConst( lexsym_t vartype, double value )
{
	val = new VarValue( value );
	type = vartype;
}


DoubleVarOrConst::DoubleVarOrConst( lexsym_t vartype, char *value )
{
	val = new VarValue( value );
	type = vartype;
}

DoubleVarOrConst::~DoubleVarOrConst()
{
}


/* get leaf value */
VarValue *
DoubleVarOrConst::getValue()
{
	return val;
}


